module OceanDynamo 
  module Associations 
    
    #
    # CollectionAssociation is an abstract class that provides common stuff to
    # ease the implementation of association proxies that represent
    # collections. See the class hierarchy in AssociationProxy.
    #
    #  Associations
    #    Association
    #      CollectionAssociation:
    #        HasAndBelongsToManyAssociation => has_and_belongs_to_many
    #        HasManyAssociation => has_many
    #          HasManyThroughAssociation + ThroughAssociation => has_many :through
    #
    # CollectionAssociation class provides common methods to the collections
    # defined by +has_and_belongs_to_many+, +has_many+ or +has_many+ with
    # +:through association+ option.
    #
    # You need to be careful with assumptions regarding the target: The proxy
    # does not fetch records from the database until it needs them, but new
    # ones created with +build+ are added to the target. So, the target may be
    # non-empty and still lack children waiting to be read from the database.
    # If you look directly to the database you cannot assume that's the entire
    # collection because new records may have been added to the target, etc.
    #
    # If you need to work on all current children, new and existing records,
    # +load_target+ and the +loaded+ flag are your friends.
    #
    class CollectionAssociation < Association 

      #
      # Resets the collection to the empty array.
      #
      def reset
        super
        @target = []
      end

    end
  end
end
